// E3D_Model.cpp: implementation of the E3D_Model class.
//
//////////////////////////////////////////////////////////////////////
#include "windows.h"
#include "stdio.h"
#include "E3D.h"
#include "E3D_Model.h"

int debug = 0;
char mystring[256];

extern "C" void cross_product_f(float x1, float y1, float z1, 
							float x2, float y2, float z2, 
							float *xout, float *yout, float *zout);

extern "C" void normalize_vector_f(float *x, float *y, float *z);

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

E3D_Model::E3D_Model()
{

}

E3D_Model::~E3D_Model()
{

}

// populate this 3D model from a file
int E3D_Model::Load(char *filename)
{
	if(Load3DO(filename)) {
		sprintf(mystring, "Error loading 3D Model file %s !!!", filename);
		MessageBox(NULL, mystring, "!!!",MB_YESNO|MB_ICONQUESTION);
		return 1;
	}
	return 0;
}

// **********************************************************************
// Reset model data to empty model (called before model load)
// **********************************************************************
void E3D_Model::Reset() {
	int i, j;
    numverts = 0;
    numpolys = 0;

    // Really only important to reset poly data (colours etc)
    for(i=0; i < MAX_POLYS_PER_MODEL; i++) {
    	poly[i].type = 0;
    	poly[i].shade = 0;
    	poly[i].gouraud = 0;
    	poly[i].filling = 0;
   		poly[i].colour.r = 0.0;
   		poly[i].colour.g = 0.0;
   		poly[i].colour.b = 0.0;
        for(j=0; j < 6; j++) {
    		poly[i].vcol[j].r = 0.0;
    		poly[i].vcol[j].g = 0.0;
    		poly[i].vcol[j].b = 0.0;
		}
	}	// next i

}

// Load in a 3DO model
// note: fix return values
int E3D_Model::Load3DO(char *filename) {

	int validcmd, validparam, end, n, i, j, v;
	float x, y, z, r, g, b;
	char scmd[80], stemp[80];
	FILE *pfile;

	fprintf(stderr, "Opening model file %s\n", filename);
	pfile = fopen(filename, "r");
	// check if file open failed
	if(pfile == NULL) {
		fprintf(stderr, "Model file open failed.\n");
        return(5);
	}

	// Reset model values
    Reset();

	// Parse model file
	end = FALSE;
	numverts = 0;
    numnorms = 0;
	numpolys = 0;
	//modelnum = myscene.nummodels;         !legacy!
	while(!end) {
		// reset valid flags to trap errors
		validcmd = FALSE; validparam = FALSE;

		fscanf(pfile, "%s", &scmd[0]);
		if(debug) fprintf(stderr, "Command is %s\n", scmd);
		if(scmd[0] == '!') {
			validcmd = TRUE;
			if(debug) fprintf(stderr, "Comment encountered. Scanning till eol.\n");
			// read till end of line (character 10)
			fscanf(pfile, "%c", &stemp[0]);
			while(stemp[0] != 10) fscanf(pfile, "%c", &stemp[0]);
		}
		if(strcmp(scmd, "vert") == 0) {
        	if(numverts == MAX_VERTS_PER_MODEL - 1 ) {
        		fprintf(stderr, "Error - too many verts! Exiting...\n");
				fclose(pfile);
				return(6);		// error
			}
			validcmd = TRUE;
			fscanf(pfile, "%f %f %f",
				&x, &y, &z);
			if(debug) fprintf(stderr, "vert: xyz %3.3f %3.3f %3.3f\n",
				x, y, z);
			vert[numverts].x = x;
			vert[numverts].y = y;
			vert[numverts].z = z;
            // verts default colour
			vert[numverts].r = 0.5;
			vert[numverts].g = 0.5;
			vert[numverts].b = 0.5;
			numverts++;
		}

		if(strcmp(scmd, "norm") == 0) {
        	if(numnorms == MAX_VERTS_PER_MODEL - 1 ) {
        		fprintf(stderr, "Error - too many normals! Exiting...\n");
				fclose(pfile);
				return(6);		// error
			}
			validcmd = TRUE;
			fscanf(pfile, "%f %f %f",
				&x, &y, &z);
			if(debug) fprintf(stderr, "norm: xyz %3.3f %3.3f %3.3f\n",
				x, y, z);
			norm[numnorms].x = x;
			norm[numnorms].y = y;
			norm[numnorms].z = z;
			numnorms++;
		}

		if(strcmp(scmd, "poly") == 0) {
        	if(numpolys == MAX_POLYS_PER_MODEL - 1 ) {
        		fprintf(stderr, "Error - too many polys! Exiting...\n");
				fclose(pfile);
				return(7);		// error
			}
			validcmd = TRUE;
            // set poly default values
			poly[numpolys].shade = 0;
			poly[numpolys].gouraud = 0;
            poly[numpolys].filling = 0;
            poly[numpolys].colour.r = 0.5;
            poly[numpolys].colour.g = 0.5;
            poly[numpolys].colour.b = 0.5;
            // one normal per poly, for now...
            poly[numpolys].n[0] = numpolys;
			// get no of verts
			fscanf(pfile, "%d", &n);
			if(debug) fprintf(stderr, "poly: %d verts   ", n);
			poly[numpolys].numverts = n;
			// read in vert indices
			for(i=0; i<n; i++) {
				fscanf(pfile, "%d", &v);
				fprintf(stderr, "%d ", v);
				poly[numpolys].p[i] = v-1;
			}
			// read in texture name
			fscanf(pfile, "%s", &stemp[0]);
			if(debug) fprintf(stderr, "%s\n", stemp);
			// set texture index here
			numpolys++;
		}

        // colour for poly and for verts
		if(strcmp(scmd, "colour") == 0) {
			validcmd = TRUE;
            j = numpolys - 1;
			fscanf(pfile, "%d %f %f %f",
				&n, &r, &g, &b);
			if(debug) fprintf(stderr, "poly %d colour %d   %3.3f %3.3f %3.3f\n",
				j, n, r, g, b);
			if(n == 0) {
				poly[j].colour.r = r;
				poly[j].colour.g = g;
				poly[j].colour.b = b;
				poly[j].vcol[0].r = r;
				poly[j].vcol[0].g = g;
				poly[j].vcol[0].b = b;
			}
            else {
            	if(n <= poly[j].numverts) {
					poly[j].vcol[n-1].r = r;
					poly[j].vcol[n-1].g = g;
					poly[j].vcol[n-1].b = b;
				}
                else {
			    	fprintf(stderr, "Error - vertex index exceeds poly numverts! Exiting...\n");
					fclose(pfile);
					return(2);		// error
				}
			}
		}

        // texture map ("texture page") index
		if(strcmp(scmd, "texture") == 0) {
			validcmd = TRUE;
            j = numpolys - 1;
			fscanf(pfile, "%d", &n);
			if(debug) fprintf(stderr, "poly %d texture %d\n", j, n);
            	if((n > -1) && (n < 16)) {
					poly[j].texmap = n;
				}
                else {
			    	fprintf(stderr, "Error - texture map index out of bounds! Exiting...\n");
					fclose(pfile);
					return(2);		// error
				}
		}

        // texture co-ordinates
		if(strcmp(scmd, "texvert") == 0) {
			validcmd = TRUE;
            j = numpolys - 1;
			fscanf(pfile, "%d %f %f",
				&n, &x, &y);
			if(debug) fprintf(stderr, "poly %d texvert %d   %3.1f %3.1f\n",
				j, n, x, y);
            	if(n <= poly[j].numverts) {
					poly[j].texp[n].x = x;
					poly[j].texp[n].y = y;
				}
                else {
			    	fprintf(stderr, "Error - texture vertex index exceeds poly numverts! Exiting...\n");
					fclose(pfile);
					return(2);		// error
				}
		}

		if(strcmp(scmd, "filling") == 0) {
			validcmd = TRUE;
            j = numpolys - 1;
			fscanf(pfile, "%d",
				&n);
			if(debug) fprintf(stderr, "poly %d filling %d\n",
				j, n);
			poly[j].filling = n;
		}

		if(strcmp(scmd, "shade") == 0) {
			validcmd = TRUE;
            j = numpolys - 1;
			fscanf(pfile, "%d",
				&n);
			if(debug) fprintf(stderr, "poly %d shade %d\n",
				j, n);
			poly[j].shade = n;
		}

		if(strcmp(scmd, "gouraud") == 0) {
			validcmd = TRUE;
            j = numpolys - 1;
			fscanf(pfile, "%d",
				&n);
			if(debug) fprintf(stderr, "poly %d gouraud %d\n",
				j, n);
			poly[j].gouraud = n;
		}

		if(strcmp(scmd, "end") == 0) {
			validcmd = TRUE;
			fprintf(stderr, "End of model file.\n");
			end = TRUE;
			// set model parameters
			numverts = numverts;
            numnorms = numnorms;
			numpolys = numpolys;
			strcpy(name, filename);
			if(debug) fprintf(stderr, "%d verts and %d polys loaded.\n", numverts, numpolys);
		}

		// check if no valid command found (ie: parse error)
		if(!validcmd) {
			fprintf(stderr, "Error - Model Command not recognised! Exiting...\n");
			fclose(pfile);
			return(2);		// error
		}

  // check if EOF before end
  if(feof(pfile) && !end) {
			fprintf(stderr, "Error - EOF before end marker! Exiting...\n");
			fclose(pfile);
			return(3);		// error
		}

	}

	fclose(pfile);

	CalculateNormals();
	return(0);		// success!
}

// **********************************************************************
// Calculate Normals for model and etc
// **********************************************************************
// (amend to calc poly and vertex normals later)
int E3D_Model::CalculateNormals() {
	int i, p1, p2, p3;
    Vector_f tnorm;
	FILE *pfile;

	//pfile = fopen(".\\log.log", "w");

	//fprintf(pfile, "Calculating normals...\n");

	for(i=0; i < numpolys; i++) {
		p1 = poly[i].p[0];
		p2 = poly[i].p[1];
		p3 = poly[i].p[2];
		cross_product_f(vert[p2].x - vert[p1].x,
            			vert[p2].y - vert[p1].y,
                        vert[p2].z - vert[p1].z,
                        vert[p2].x - vert[p3].x,
                        vert[p2].y - vert[p3].y,
                        vert[p2].z - vert[p3].z,
                        &tnorm.x, &tnorm.y, &tnorm.z);
		normalize_vector_f(&tnorm.x, &tnorm.y, &tnorm.z);
        // write model normals
        norm[i].x = tnorm.x;
        norm[i].y = tnorm.y;
        norm[i].z = tnorm.z;
        // set poly normal index
        poly[i].n[0] = i;
	}
    numnorms = i;
	//fprintf(pfile, "%d normals calculated.\n", i);
	//fclose(pfile);
	return(i);
}

